package cz.slabi.david.animaletto;

import java.awt.Image;
import java.awt.Toolkit;

/**
 * Acquires and keeps information about cards.
 * 
 * @author DavidSlaby
 */
public class Card {
	
    /**
     * Types of cards used in the game.
     */
	public enum CardType { COLOR, JOKER, BONUS, ROW, LAST, BACK }
	
    /**
     * Number of colors of cards.
     */
    public static final int allColorCount = 7;
    
    /**
     * Image of wood for the table.
     */
    public static Image table;
    
    /**
     * Images of color cards.
     */
    public static Image[] colorImage = new Image[allColorCount];
    
    /**
     * Images of chameleon, bonus, row, last round and card back.
     */
    public static Image[] otherImage = new Image[CardType.values().length];
    
    /**
     * Loads all images used in the game.
     */
    public static void loadImages(){
		Toolkit tk = Toolkit.getDefaultToolkit();
		
		table = tk.getImage("data/img/wood.jpg");
        for (int i = 0; i < allColorCount; i++){
            colorImage[i] = tk.getImage("data/img/card"+String.valueOf(i)+".png");
        }
        otherImage[1] = tk.getImage("data/img/joker.png");
        otherImage[2] = tk.getImage("data/img/bonus.png");
        otherImage[3] = tk.getImage("data/img/row.png");
        otherImage[4] = tk.getImage("data/img/last.png");
        otherImage[5] = tk.getImage("data/img/back.png");

    }

    public CardType type;
    public int color;

    /**
     * Creates a card of specific type and color.
     * 
     * @param type      type of the card
     * @param color     color of the card (meaningful with type COLOR)
     */
    Card(CardType type, int color){
        this.type = type;
        this.color = color;
    }
    
    /**
     * Creates a card of specific type. Should not be color.
     * 
     * @param type      type of the card
     */
    Card(CardType type){
        this(type, -1);
    }
	
    /**
     * Returns image associated with the card according to its type and color.
     * 
     * @return  image of the card
     */
    public Image getCardImage(){
        switch (type){
			case COLOR:
				return colorImage[color];
			case JOKER:
				return otherImage[1];
			case BONUS:
				return otherImage[2];
			case ROW:
				return otherImage[3];
			case LAST:
				return otherImage[4];
			default:
				return null;
		}
    }
    

}
