package cz.slabi.david.animaletto;

import java.util.LinkedList;
import java.util.Random;

/**
 * The card deck used during the game. Should be created pseudo-randomly with
 * a seed common to all players.
 * 
 * @author DavidSlaby
 */
public class Deck extends LinkedList<Card>{
    
    public int colorCount;
    
    /**
     * Creates a deck according to number of players and randomizer.
     * 
     * @param randomizer    a randomizer used to shake cards
     * @param playercnt     number of players (defines the amount of colors used)
     */
    public Deck(Random randomizer, int playercnt){
		int count;
        
        colorCount = (Main.client.playersCount < 4)
                ? Card.allColorCount - 4 + Main.client.playersCount
                : Card.allColorCount;
        
        LinkedList<Card> temp = new LinkedList<Card>();
		
		for (int j=0; j < colorCount; j++){
			count = (j < playercnt) ? 8 : 9;
			for (int i = 0; i < count; i++) temp.add(new Card(Card.CardType.COLOR, j));
		}
		for (int i = 0; i < 3; i++) temp.add(new Card(Card.CardType.JOKER));
		for (int i = 0; i < 10; i++) temp.add(new Card(Card.CardType.BONUS));
		
		while (!temp.isEmpty()){
			int rnd = randomizer.nextInt(temp.size());
			add(temp.get(rnd));
			temp.remove(rnd);
		}
		add(size() - 12, new Card(Card.CardType.LAST));
	}

}
