package cz.slabi.david.animaletto;

import java.awt.*;
/**
 * Main window of the application. Contains all the information the user - player
 * needs.
 * 
 * @author  DavidSlaby
 */
public class MainWindow extends javax.swing.JFrame {

    /**
     * The desk of the table where the game is played. Cards are put on it.
     * Is double-buffered.
     */
	public class MyCanvas extends java.awt.Canvas{
	
		private int startPos;
		private Point center = new Point();
		private int minDimension;
		private double packetCenterDistance;
		private double ownedCenterDistance;
		private int cardWidth;
		private int cardHeight;
		
	    private int bufferWidth;
	    private int bufferHeight;
		private Image bufferImage;
		private Graphics bufferGraphics;

		private Rectangle recDeck;
		private Rectangle recLast;
		private Rectangle recPicked;
		private Rectangle[] recPackets;
		private Rectangle[] recWholePackets;
		private Point[] packetShiftVector;
		
		private Rectangle[] recOwned;
		
		private Point[] ownedMiddle;
		private Point[] ownedShiftVector;
		private Point[] ownedStoreLine;
		
		private Graphics graphics;
        
        /**
         * Creates the table.
         */
		MyCanvas(){
			resized();
		}

        /**
         * Tests whether second buffer needs to reset.
         */
		private void testBuffer(){
			if(bufferWidth!=getSize().width || 
				bufferHeight!=getSize().height || 
				bufferImage==null || bufferGraphics==null)
				resetBuffer();
		}
		
        /**
         * Resets the second buffer.
         */
        private void resetBuffer(){
			bufferWidth=getSize().width;
			bufferHeight=getSize().height;

			if(bufferGraphics!=null){
				bufferGraphics.dispose();
				bufferGraphics=null;
			}
			if(bufferImage!=null){
				bufferImage.flush();
				bufferImage=null;
			}
			System.gc();

			bufferImage=createImage(bufferWidth,bufferHeight);
			bufferGraphics=bufferImage.getGraphics();
		}

        /**
         * Calls painting onto second buffer and then puts it on the canvas.
         * @param g     the graphics contained by the canvas
         */
        @Override
        public void paint(Graphics g){

			testBuffer();
			
			if(bufferGraphics!=null){
				bufferGraphics.clearRect(0,0,bufferWidth,bufferHeight);
				paintBuffer();
				g.drawImage(bufferImage,0,0,this);
	        }
        }

        /**
         * Paints the wooden background and cards onto the second buffer.
         */
        private void paintBuffer(){
			graphics = bufferGraphics;

			graphics.drawImage(Card.table, 0, 0, table.getWidth(), table.getHeight(), this);
			tableRefreshDraw();
		}

        /**
         * Simple override of update to avoid blinking.
         * 
         * @param g     used graphics
         */
		@Override
		public void update(Graphics g){
			paint(g);
		}
		
        /**
         * Refreshes the contents of the table.
         * 
         * @param resize    if true, positions of cards are counted again
         */
		public void tableRefresh(boolean resize){
			if (resize) resized();
			tableRefresh();
		}

        /**
         * Refreshes the contents of the table.
         */
		private void tableRefresh(){
			repaint();
		}
		
        /**
         * Puts Cards onto the table.
         */
		private void tableRefreshDraw(){
            
            if (Main.client.deck == null) return;
            
			int playerCount = Main.client.playersCount;
			
			putPacketCards(playerCount);
			
			putOwnedCards(playerCount);
			
			if (Main.client.lastRound)
				table.graphics.drawImage(Card.otherImage[4], recLast.x, recLast.y, 
					recLast.width, recLast.height, this);

			if (!Main.client.deck.isEmpty()) 
				table.graphics.drawImage(Card.otherImage[5], recDeck.x, recDeck.y, 
					recDeck.width, recDeck.height, this);
			
			if (Main.client.picked != null)
				table.graphics.drawImage(Main.client.picked.getCardImage(), recPicked.x, recPicked.y, 
				recPicked.width, recPicked.height, this);
		}
		
        /**
         * Puts packets to the middle of the table.
         * 
         * @param playerCount   number of players is equal to the number of packets
         */
		private void putPacketCards(int playerCount){
			Card myCard;
			for (int i = 0; i < playerCount; i++){
				if (Main.client.packets[i] == null) continue;
				for (int j = 0; j < 3; j++){
					myCard = Main.client.packets[i][j];
					if ((myCard != null) || (j == 0)){
					    if (myCard == null) myCard = new Card(Card.CardType.ROW);
						Image img;
						img = (myCard != null) ? myCard.getCardImage() : Card.otherImage[3];
						graphics.drawImage(img,
							recPackets[i].x + packetShiftVector[i].x * j,
							recPackets[i].y + packetShiftVector[i].y * j, 
							recPackets[i].width, recPackets[i].height, this);
					}
					if (myCard == null) break;
				}
			}
		}

        /**
         * Puts owned cards 'in front of' players.
         * 
         * @param playerCount   number of players
         */
		private void putOwnedCards(int playerCount){
			
			Card myCard;
			
			for (int i = 0; i < playerCount; i++){
				int ccTypes = 0;
				int[] cardColors = new int[Main.client.deck.colorCount + 2];
				int[] jokers = new int[Main.client.deck.colorCount + 2];
				for (int j = 0; j < Main.client.ownedCount[i]; j++){
					myCard = Main.client.owned[i][j];
					if (myCard.color >= 0){
						if (cardColors[myCard.color] == 0) ccTypes++;
						if (myCard.type == Card.CardType.JOKER) jokers[myCard.color]++;
						cardColors[myCard.color]++;
					}else if (myCard.type == Card.CardType.JOKER){
						if (cardColors[Main.client.deck.colorCount] == 0) ccTypes++;
						cardColors[Main.client.deck.colorCount]++;
					}else if (myCard.type == Card.CardType.BONUS){
						if (cardColors[Main.client.deck.colorCount + 1] == 0) ccTypes++;
						cardColors[Main.client.deck.colorCount + 1]++;
					}
				}
				
				int curColor = 0; // ordinal number of my packet
				for (int j = 0; j < Main.client.deck.colorCount+2; j++){
					
					if (cardColors[j] == 0) continue; // no cards - no packet
					curColor++;
					
					if (j < Main.client.deck.colorCount) myCard = new Card(Card.CardType.COLOR, j);
					else if (j == Main.client.deck.colorCount) myCard = new Card(Card.CardType.JOKER);
					else if (j == Main.client.deck.colorCount + 1) myCard = new Card(Card.CardType.BONUS);
					else myCard = new Card(Card.CardType.BACK);
					
					Image img = myCard.getCardImage();
					
					double groupDistanceX = (double)ownedStoreLine[i].x / (ccTypes + 1);
					double groupDistanceY = (double)ownedStoreLine[i].y / (ccTypes + 1);
					int cardCenterX = ownedMiddle[i].x - ownedStoreLine[i].x / 2 +
						(int)Math.round(groupDistanceX * curColor);
					int cardCenterY = ownedMiddle[i].y - ownedStoreLine[i].y / 2 +
						(int)Math.round(groupDistanceY * curColor);
					
					for (int ii = 0; ii < cardColors[j]; ii++){
						if (ii == cardColors[j] - jokers[j]) img = Card.otherImage[1];
						Rectangle recCard = new Rectangle
							(cardCenterX - cardWidth / 2, cardCenterY - cardWidth / 2, cardWidth, cardHeight);
						graphics.drawImage(img,
							recCard.x + ownedShiftVector[i].x * ii,
							recCard.y + ownedShiftVector[i].y * ii,
							cardWidth, cardHeight, this);
						if ((i == Main.client.myPosition) && (ii == 0) && (j < Main.client.deck.colorCount)){
							recOwned[j] = recCard;
							recOwned[j].height += cardColors[j] * ownedShiftVector[i].y;
						}
					}
				}
			}
		}
		
        /**
         * Counts again all the locations of cards.
         */
		private void resized(){

			if ((Main.client == null) 
                || (Main.client.deck == null)) return;

            int playerCount = Main.client.playersCount;

			recDeck = null;
			recLast = null;
			recPackets = null;
			packetShiftVector = null;

			recOwned = new Rectangle[Main.client.deck.colorCount];

			startPos = Main.client.myPosition;
			
			center.x = table.getWidth() / 2;
			center.y = table.getHeight() / 2;
			
			minDimension = Math.min(table.getHeight(), table.getWidth());
			
			packetCenterDistance = minDimension / 5;
			ownedCenterDistance = minDimension / 3;

            double diminisher = 4 + playerCount;
            
			cardWidth = (int)Math.round(minDimension / diminisher * 0.5);
			cardHeight = (int)Math.round(minDimension / diminisher * 0.8);
			
			recDeck = new Rectangle(center.x - (int)Math.round(cardWidth * 1.6),
				center.y - cardHeight / 2, cardWidth, cardHeight);
			recLast = new Rectangle(center.x + (int)Math.round(cardWidth * 0.6),
				center.y - cardHeight / 2, cardWidth, cardHeight);
			recPicked = new Rectangle(center.x - cardWidth / 2,
				center.y - cardHeight / 2, cardWidth, cardHeight);

			recPackets = new Rectangle[playerCount];
			recWholePackets = new Rectangle[playerCount];
			packetShiftVector = new Point[playerCount];
			
			ownedMiddle = new Point[playerCount];
			ownedShiftVector = new Point[playerCount];
			ownedStoreLine = new Point[playerCount];

			double neighbourAngle = 2 * Math.PI / playerCount;
			double orderShift = cardHeight * 0.25;
			
			for (int i = 0; i < playerCount; i++){
				
				double anglePacket = Math.PI + (i - startPos) * neighbourAngle + neighbourAngle / 2;
				double angleOwned = Math.PI + (i - startPos) * neighbourAngle;

				int xPacketCenter = 
					(int)Math.round(center.x + packetCenterDistance * Math.sin(anglePacket));
				int yPacketCenter = 
					(int)Math.round(center.y + packetCenterDistance * (-Math.cos(anglePacket)));
				
				int xCorner = xPacketCenter - cardWidth / 2;
				int yCorner = yPacketCenter - cardHeight / 2;
				
				ownedMiddle[i] = new Point
					((int)Math.round(center.x + ownedCenterDistance * Math.sin(angleOwned)),
					(int)Math.round(center.y + ownedCenterDistance * (-Math.cos(angleOwned))));
				
				recPackets[i] = new Rectangle(xCorner, yCorner, cardWidth, cardHeight);

				packetShiftVector[i] = new Point((int)Math.round(orderShift * Math.sin(anglePacket)),
					(int)Math.round(orderShift * (-Math.cos(anglePacket))));

				ownedShiftVector[i] = new Point((int)Math.round(orderShift * Math.sin(angleOwned)),
					(int)Math.round(orderShift * (-Math.cos(angleOwned))));
				ownedStoreLine[i] = new Point(ownedShiftVector[i].y * 24, -ownedShiftVector[i].x * 24);

				recWholePackets[i] = (Rectangle)recPackets[i].clone();
				if (packetShiftVector[i].x < 0)
					recWholePackets[i].x += packetShiftVector[i].x * 2;
				if (packetShiftVector[i].y < 0)
					recWholePackets[i].y += packetShiftVector[i].y * 2;
				recWholePackets[i].width += Math.abs(packetShiftVector[i].x) * 2;
				recWholePackets[i].height += Math.abs(packetShiftVector[i].y) * 2;
			}
			repaint();
		}
	}
	
    /**
     * Manages selecting a card or packet during the game.
     * 
     * @param evt   event of mouse clicking onto the table
     */
	private void tableMouseClicked(java.awt.event.MouseEvent evt) {
		
		if (Main.client.gameState == Client.GameState.PLAY){

		    if ((Main.client.cardAction == Client.CardAction.NONE) &&
			    (!Main.client.deck.isEmpty()) &&
			    (table.recDeck.contains(evt.getPoint()))){
			    Main.client.doPlayCardAction(-1);
		    }

		    int playerCount = Main.client.playersCount;

		    if (Main.client.cardAction == Client.CardAction.NONE){
				for (int i = 0; i < playerCount; i++){
				    if ((Main.client.packets[i] == null) || (Main.client.packets[i][0] == null)) continue;
				    if (table.recWholePackets[i].contains(evt.getPoint())){
					    Main.client.doPlayCardAction(i);
					    break;
				    }
				}
		    }else if (Main.client.cardAction == Client.CardAction.PICKCARD){
				for (int i = 0; i < playerCount; i++){
				    if (table.recWholePackets[i].contains(evt.getPoint())){
					    Main.client.doPlayCardAction(i);
					    break;
				    }
				}
		    }
		}else if (Main.client.gameState == Client.GameState.FINISH){
            
            if ((table.recLast.contains(evt.getPoint()))){
			    Main.client.doFinishCardAction(-2);
		    }
            
		    for (int i = 0; i < Main.client.deck.colorCount; i++){
			    
			    if ((table.recOwned[i] != null) && (table.recOwned[i].contains(evt.getPoint()))){
				    Main.client.doFinishCardAction(i);
				    break;
			    }
			}
		}
		
		table.repaint();
	}
	
	/** 
     * Creates the main window.
     */
	public MainWindow() {
	    setTitle("Animaletto Card Game");
		initComponents();
        table = new MyCanvas();
        panelTable.add(table, java.awt.BorderLayout.CENTER);
        getContentPane().add(panelTable, java.awt.BorderLayout.CENTER);
        table.setBackground(new java.awt.Color(96, 48, 0));
        table.addMouseListener(new java.awt.event.MouseAdapter() {
			@Override
            public void mouseClicked(java.awt.event.MouseEvent evt) {
                tableMouseClicked(evt);
            }
        });
        setSize(640, 480);
		setVisible(true);
		table.repaint();
	}
	
	/** 
     * Initializes the components and their event-handlers.
	 */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        panelFields = new javax.swing.JPanel();
        chatField = new javax.swing.JTextField();
        scrollChatArea = new javax.swing.JScrollPane();
        chatArea = new javax.swing.JTextArea();
        scrollLog = new javax.swing.JScrollPane();
        log = new javax.swing.JTextArea();
        panelTable = new javax.swing.JPanel();
        labelHint = new java.awt.Label();
        menu = new javax.swing.JMenuBar();
        menuGame = new javax.swing.JMenu();
        menuSetGame = new javax.swing.JMenuItem();
        menuReady = new javax.swing.JCheckBoxMenuItem();
        menuQuit = new javax.swing.JMenuItem();
        menuNetwork = new javax.swing.JMenu();
        menuConnect = new javax.swing.JMenuItem();
        menuNetSet = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setLocationByPlatform(true);
        setMinimumSize(new java.awt.Dimension(530, 320));
        addComponentListener(new java.awt.event.ComponentAdapter() {
            public void componentResized(java.awt.event.ComponentEvent evt) {
                formComponentResized(evt);
            }
        });

        panelFields.setPreferredSize(new java.awt.Dimension(165, 209));
        panelFields.setRequestFocusEnabled(false);
        panelFields.setLayout(new java.awt.BorderLayout(2, 2));

        chatField.setEnabled(false);
        chatField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                chatFieldKeyReleased(evt);
            }
        });
        panelFields.add(chatField, java.awt.BorderLayout.NORTH);

        chatArea.setColumns(20);
        chatArea.setLineWrap(true);
        chatArea.setRows(5);
        chatArea.setWrapStyleWord(true);
        chatArea.setEnabled(false);
        chatArea.setFocusable(false);
        scrollChatArea.setViewportView(chatArea);

        panelFields.add(scrollChatArea, java.awt.BorderLayout.CENTER);

        scrollLog.setMinimumSize(new java.awt.Dimension(22, 100));
        scrollLog.setPreferredSize(new java.awt.Dimension(203, 120));

        log.setColumns(20);
        log.setLineWrap(true);
        log.setRows(5);
        log.setFocusable(false);
        scrollLog.setViewportView(log);

        panelFields.add(scrollLog, java.awt.BorderLayout.PAGE_END);

        getContentPane().add(panelFields, java.awt.BorderLayout.WEST);

        panelTable.setBorder(javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.LOWERED));
        panelTable.setLayout(new java.awt.BorderLayout(4, 4));

        labelHint.setAlignment(java.awt.Label.CENTER);
        labelHint.setBackground(new java.awt.Color(102, 51, 0));
        labelHint.setFont(new java.awt.Font("DejaVu Sans", 1, 14));
        labelHint.setForeground(new java.awt.Color(255, 255, 204));
        labelHint.setText("void");
        panelTable.add(labelHint, java.awt.BorderLayout.NORTH);

        getContentPane().add(panelTable, java.awt.BorderLayout.CENTER);

        menuGame.setText("Game");

        menuSetGame.setText("Settings...");
        menuSetGame.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuSetGameActionPerformed(evt);
            }
        });
        menuGame.add(menuSetGame);

        menuReady.setText("Ready");
        menuReady.setEnabled(false);
        menuReady.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuReadyActionPerformed(evt);
            }
        });
        menuGame.add(menuReady);

        menuQuit.setText("Quit");
        menuQuit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuQuitActionPerformed(evt);
            }
        });
        menuGame.add(menuQuit);

        menu.add(menuGame);

        menuNetwork.setText("Network");

        menuConnect.setText("Connect");
        menuConnect.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuConnectActionPerformed(evt);
            }
        });
        menuNetwork.add(menuConnect);

        menuNetSet.setText("Settings...");
        menuNetSet.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuNetSetActionPerformed(evt);
            }
        });
        menuNetwork.add(menuNetSet);

        menu.add(menuNetwork);

        setJMenuBar(menu);

        pack();
    }// </editor-fold>//GEN-END:initComponents
	
    /**
    * Refreshes the contents of the table.
    * 
    * @param resize    if true, positions of cards are counted again
    */
	public void tableRefresh(boolean resize){
		table.tableRefresh(resize);
	}
	
    /**
     * Puts next line into the log.
     * 
     * @param record    the new line being put to the log
     */
	public void logRecord(String record){
		log.setText(log.getText()+record+"\n");
	}
	
    /**
     * Puts next line into the chat.
     * 
     * @param name      name of the sender of the message
     * @param record    the new line being put to the chat
     */
	public void chatRecord(String name, String record){
		chatArea.setText(chatArea.getText()+name+": "+record+"\n");
	}
	
    /**
     * Changes in the GUI after succesful connection to the server.
     */
	public void gotConnected(){
		chatField.setEnabled(true);
		chatArea.setEnabled(true);
	}
	
    /**
     * Prepares GUI for the start of the game.
     */
	public void prepareGame(){
		table.resized();
		menuReady.setEnabled(false);
		this.logRecord("Everyone is ready.");
	}
    
    /**
     * Sets the content of the Hint label.
     * @param text  Content of the label.
     */
    public void setHint(String text){
        if (!text.equals("")) labelHint.setText(text);
    }
    
    /**
     * Prepares GUI for end of games.
     * 
     * @param alright   true when all games finished, false when someone disconnected
     */
    public void finishedGames(boolean alright){
        menuReady.setState(false);
        menuReady.setEnabled(true);
        
        if (alright)
            logRecord("Game ended.");
        else
            logRecord("A player has disconnected, game ended.");
		
		tableRefresh(false);
    }
	
    /**
     * Connects to or disconnects from the server and sets GUI.
     */
	public void disConnect(){
		if (Main.client.socket.isClosed()){
			try{
				Main.client.start();
				menuConnect.setText("Disconnect");
				menuReady.setEnabled(true);
				menuSetGame.setEnabled(false);
				menuNetSet.setEnabled(false);
			}catch (Exception e){
				javax.swing.JOptionPane.showMessageDialog(this, "Network was not set properly", 
					"Invalid Network Settings", javax.swing.JOptionPane.ERROR_MESSAGE);
			}
		}else{
			Main.client.stop();
			menuReady.setState(false);
			menuReady.setEnabled(false);
			menuSetGame.setEnabled(true);
			menuNetSet.setEnabled(true);
			chatField.setEnabled(false);
			chatArea.setEnabled(false);
			menuConnect.setText("Connect");
			tableRefresh(false);
		}
	}

    /** 
     * (Dis)connecting.
     * @param evt   the corresponding event
     */
	private void menuConnectActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuConnectActionPerformed
		disConnect();
	}//GEN-LAST:event_menuConnectActionPerformed

    /**
     * Showing the network settings dialog.
     * @param evt   the corresponding event
     */
	private void menuNetSetActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuNetSetActionPerformed
		new NetworkDialog(new java.awt.Frame(), true);
	}//GEN-LAST:event_menuNetSetActionPerformed

    /**
     * Showing the game settings dialog.
     * @param evt   the corresponding event
     */
	private void menuSetGameActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuSetGameActionPerformed
		new GameDialog(new java.awt.Frame(), true);
	}//GEN-LAST:event_menuSetGameActionPerformed

    /**
     * Exiting the aplication with clean disconnecting.
     * @param evt   the corresponding event
     */
	private void menuQuitActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuQuitActionPerformed
		System.exit(0);
	}//GEN-LAST:event_menuQuitActionPerformed

    /**
     * Sending a chat message.
     * @param evt   the corresponding event
     */
	private void chatFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_chatFieldKeyReleased
		if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_ENTER){
			Main.client.chat(chatField.getText());
			chatRecord(Main.client.name, chatField.getText());
			chatField.setText("");
		}
	}//GEN-LAST:event_chatFieldKeyReleased

    /**
     * Setting the Ready button.
     * 
     * @param evt
     */
	private void menuReadyActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuReadyActionPerformed
		Main.client.isReady = menuReady.getState();
	}//GEN-LAST:event_menuReadyActionPerformed

    /**
     * Setting table stuff location according to the new size.
     * 
     * @param evt   the corresponding event
     */
	private void formComponentResized(java.awt.event.ComponentEvent evt) {//GEN-FIRST:event_formComponentResized
		// TODO add your handling code here:
		if (table != null) table.resized();
	}//GEN-LAST:event_formComponentResized
	
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JTextArea chatArea;
    private javax.swing.JTextField chatField;
    private java.awt.Label labelHint;
    private javax.swing.JTextArea log;
    private javax.swing.JMenuBar menu;
    private javax.swing.JMenuItem menuConnect;
    private javax.swing.JMenu menuGame;
    private javax.swing.JMenuItem menuNetSet;
    private javax.swing.JMenu menuNetwork;
    private javax.swing.JMenuItem menuQuit;
    private javax.swing.JCheckBoxMenuItem menuReady;
    private javax.swing.JMenuItem menuSetGame;
    private javax.swing.JPanel panelFields;
    private javax.swing.JPanel panelTable;
    private javax.swing.JScrollPane scrollChatArea;
    private javax.swing.JScrollPane scrollLog;
    // End of variables declaration//GEN-END:variables
	
    private MyCanvas table;
}
